import pytest
from django.core.exceptions import ObjectDoesNotExist
from django.urls import reverse
from django_quotes.models import Quote, Source, SourceGroup

from tests.factories.users import UserFactory

pytestmark = pytest.mark.django_db(transaction=True)

# FIXTURES


@pytest.fixture
def c_groups_user(user):
    """

    :param user: A user object from the user fixture
    :return: A list of created groups to use in tests
    """
    user2 = UserFactory()
    group_list = []
    group_list.append(SourceGroup.objects.create(name="Group1", owner=user))
    group_list.append(SourceGroup.objects.create(name="Group2", owner=user))
    group_list.append(SourceGroup.objects.create(name="Group3", owner=user2))
    c1 = Source.objects.create(name="Johnny Boy", group=group_list[0], owner=user)
    Source.objects.create(name="Mary", group=group_list[0], owner=user)
    c2 = Source.objects.create(name="Markus", group=group_list[1], owner=user)
    Quote.objects.create(quote="Up is down.", source=c1, owner=user)
    Quote.objects.create(quote="Down is up", source=c1, owner=user)
    Quote.objects.create(quote="Bananas are a growth industry", source=c2, owner=user)
    yield user  # The user with the two groups associated with them.
    for group in group_list:
        group.delete()
    user2.delete()


# SOURCE GROUP TESTS


def test_group_list_view(client, c_groups_user, django_assert_max_num_queries):
    """
    Tests for expected results for the group list view.
    :param c_groups_user: User from the fixture
    :return:
    """
    client.force_login(user=c_groups_user)
    url = reverse("quotes:group_list")
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    assert len(response.context["groups"]) == 2


@pytest.mark.parametrize("view_name", ["quotes:group_list", "quotes:group_create"])
def test_group_requires_login(client, view_name):
    url = reverse(view_name)
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response["Location"]


def test_group_create(client, django_assert_max_num_queries, user):
    url = reverse("quotes:group_create")
    client.force_login(user)
    existing_groups = SourceGroup.objects.filter(owner=user).count()
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "John Snow", "description": "Knows nothing"})
    assert response.status_code == 302
    assert SourceGroup.objects.filter(owner=user).count() - existing_groups == 1


@pytest.mark.parametrize("view_name", ["quotes:group_detail", "quotes:group_update", "quotes:group_delete"])
def test_groups_single_object_view_requires_login(client, c_groups_user, view_name):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    url = reverse(view_name, kwargs={"group": group.slug})
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response["Location"]


def test_group_detail_view_for_owner(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    url = reverse("quotes:group_detail", kwargs={"group": group.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200


def test_group_detail_for_other_user(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    url = reverse("quotes:group_detail", kwargs={"group": group.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403


def test_group_update_access_restricted_to_owner(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    url = reverse("quotes:group_update", kwargs={"group": group.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "Doodle", "description": "I **hate** you"})
    assert response.status_code == 403
    group.refresh_from_db()
    assert group.description != "I **hate** you"


def test_group_update_by_owner(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    url = reverse("quotes:group_update", kwargs={"group": group.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "Jon Snow", "description": "Knows nothing"})
    assert response.status_code == 302
    group.refresh_from_db()
    assert group.description == "Knows nothing"


def test_group_delete_not_accessible_by_non_owner(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    group_num = SourceGroup.objects.filter(owner=c_groups_user).count()
    url = reverse("quotes:group_delete", kwargs={"group": group.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403
    with django_assert_max_num_queries(50):
        response = client.post(url, data={})
    assert response.status_code == 403
    assert group_num == SourceGroup.objects.filter(owner=c_groups_user).count()


def test_group_delete(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.filter(owner=c_groups_user)[0]
    group_num = SourceGroup.objects.filter(owner=c_groups_user).count()
    url = reverse("quotes:group_delete", kwargs={"group": group.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    with django_assert_max_num_queries(50):
        response = client.post(url, data={})
    assert response.status_code == 302
    assert group_num - SourceGroup.objects.filter(owner=c_groups_user).count() == 1


# SOURCE TESTS


@pytest.mark.parametrize("view_name", ["quotes:source_list", "quotes:source_create"])
def test_source_list_requires_login(client, c_groups_user, view_name):
    group = SourceGroup.objects.get(name="Group1")
    url = reverse(view_name, kwargs={"group": group.slug})
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response["Location"]


@pytest.mark.parametrize("view_name", ["quotes:source_list", "quotes:source_create"])
def test_unauthorized_source_list_access(client, django_assert_max_num_queries, c_groups_user, view_name):
    group = SourceGroup.objects.get(name="Group1")
    url = reverse(view_name, kwargs={"group": group.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403


def test_authorized_source_list_access(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.get(name="Group1")
    url = reverse("quotes:source_list", kwargs={"group": group.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    assert response.context["sources"].count() == 2


def test_unauthorized_source_create(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.get(name="Group1")
    source_number = Source.objects.filter(group=group).count()
    url = reverse("quotes:source_create", kwargs={"group": group.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "John Snow"})
    assert response.status_code == 403
    assert source_number == Source.objects.filter(group=group).count()


def test_authorized_source_create(client, django_assert_max_num_queries, c_groups_user):
    group = SourceGroup.objects.get(name="Group1")
    source_number = Source.objects.filter(group=group).count()
    url = reverse("quotes:source_create", kwargs={"group": group.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "Peter Parker"})
    assert response.status_code == 302
    assert Source.objects.filter(group=group).count() - source_number == 1


@pytest.mark.parametrize(
    "view_name",
    ["quotes:source_detail", "quotes:source_update", "quotes:source_delete"],
)
def test_get_views_for_source_require_login(client, c_groups_user, view_name):
    url = reverse(view_name, kwargs={"source": "group1-johnny-boy"})
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response["Location"]


@pytest.mark.parametrize(
    "view_name",
    ["quotes:source_detail", "quotes:source_update", "quotes:source_delete"],
)
def test_get_views_for_source_unauthorized(client, django_assert_max_num_queries, c_groups_user, view_name):
    url = reverse(view_name, kwargs={"source": "group1-johnny-boy"})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403


@pytest.mark.parametrize(
    "view_name",
    ["quotes:source_detail", "quotes:source_update", "quotes:source_delete"],
)
def test_get_views_source_authorized(client, django_assert_max_num_queries, c_groups_user, view_name):
    url = reverse(view_name, kwargs={"source": "group1-johnny-boy"})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200


def test_edit_source_view_unauthorized(client, django_assert_max_num_queries, c_groups_user):
    url = reverse("quotes:source_update", kwargs={"source": "group1-johnny-boy"})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "Johnny Boy", "description": "Eats worms"})
    assert response.status_code == 403
    assert Source.objects.get(slug="group1-johnny-boy").description != "Eats worms"


def test_delete_source_unauthorized(client, django_assert_max_num_queries, c_groups_user):
    url = reverse("quotes:source_delete", kwargs={"source": "group1-johnny-boy"})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.post(url, data={})
    assert response.status_code == 403
    assert Source.objects.get(slug="group1-johnny-boy")


def test_authorized_source_edit(client, django_assert_max_num_queries, c_groups_user):
    url = reverse("quotes:source_update", kwargs={"source": "group1-johnny-boy"})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"name": "Johnny Boy", "description": "Dances in the dark"})
    assert response.status_code == 302
    assert Source.objects.get(slug="group1-johnny-boy").description == "Dances in the dark"


def test_authorized_source_delete(client, django_assert_max_num_queries, c_groups_user):
    url = reverse("quotes:source_delete", kwargs={"source": "group1-johnny-boy"})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.post(url, data={})
    assert response.status_code == 302
    with pytest.raises(ObjectDoesNotExist):
        Source.objects.get(slug="group1-johnny-boy")


# QUOTES


@pytest.mark.parametrize("view_name", ["quotes:quote_list", "quotes:quote_create"])
def test_list_create_quotes_requires_login(client, c_groups_user, view_name):
    source = Source.objects.get(slug="group1-johnny-boy")
    url = reverse(view_name, kwargs={"source": source.slug})
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response["Location"]


@pytest.mark.parametrize("view_name", ["quotes:quote_list", "quotes:quote_create"])
def test_get_list_create_quotes_unauthorized(client, django_assert_max_num_queries, c_groups_user, view_name):
    source = Source.objects.get(slug="group1-johnny-boy")
    url = reverse(view_name, kwargs={"source": source.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403


def test_list_quote_authorized(client, django_assert_max_num_queries, c_groups_user):
    source = Source.objects.get(slug="group1-johnny-boy")
    url = reverse("quotes:quote_list", kwargs={"source": source.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200


def test_unauthorized_create_quote(client, django_assert_max_num_queries, c_groups_user):
    source = Source.objects.get(slug="group1-johnny-boy")
    quote_num = source.quote_set.count()
    url = reverse("quotes:quote_create", kwargs={"source": source.slug})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"quote": "I want all the ham"})
    assert response.status_code == 403
    assert quote_num == Quote.objects.filter(source=source).count()


def test_authorized_create_quote(client, django_assert_max_num_queries, c_groups_user):
    source = Source.objects.get(slug="group1-johnny-boy")
    quote_num = source.quote_set.count()
    url = reverse("quotes:quote_create", kwargs={"source": source.slug})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"quote": "I want all the ham"})
    assert response.status_code == 302
    assert Quote.objects.filter(source=source).count() - quote_num == 1


@pytest.mark.parametrize("view_name", ["quotes:quote_detail", "quotes:quote_update", "quotes:quote_delete"])
def test_get_views_quote_detail_require_login(client, c_groups_user, view_name):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse(view_name, kwargs={"quote": quote.id})
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response["Location"]


@pytest.mark.parametrize("view_name", ["quotes:quote_detail", "quotes:quote_update", "quotes:quote_delete"])
def test_get_views_quote_detail_unauthorized(client, django_assert_max_num_queries, c_groups_user, view_name):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse(view_name, kwargs={"quote": quote.id})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 403


@pytest.mark.parametrize("view_name", ["quotes:quote_detail", "quotes:quote_update", "quotes:quote_delete"])
def test_get_views_quote_detail_authorized(client, django_assert_max_num_queries, c_groups_user, view_name):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse(view_name, kwargs={"quote": quote.id})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.get(url)
    assert response.status_code == 200


def test_quote_edit_unauthorized(client, django_assert_max_num_queries, c_groups_user):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse("quotes:quote_update", kwargs={"quote": quote.id})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"quote": "I ate too much pie"})
    assert response.status_code == 403
    quote.refresh_from_db()
    assert quote.quote != "I ate too much pie"


def test_quote_edit_authorized(client, django_assert_max_num_queries, c_groups_user):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse("quotes:quote_update", kwargs={"quote": quote.id})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.post(url, data={"quote": "I ate too much pie"})
    assert response.status_code == 302
    quote.refresh_from_db()
    assert quote.quote == "I ate too much pie"


def test_quote_delete_unauthorized(client, django_assert_max_num_queries, c_groups_user):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse("quotes:quote_delete", kwargs={"quote": quote.id})
    client.force_login(UserFactory())
    with django_assert_max_num_queries(50):
        response = client.post(url, data={})
    assert response.status_code == 403
    assert Quote.objects.get(id=quote.id)


def test_quote_delete_authorized(client, django_assert_max_num_queries, c_groups_user):
    quote = Quote.objects.select_related("source").filter(source__slug="group1-johnny-boy")[0]
    url = reverse("quotes:quote_delete", kwargs={"quote": quote.id})
    client.force_login(c_groups_user)
    with django_assert_max_num_queries(50):
        response = client.post(url, data={})
    assert response.status_code == 302
    with pytest.raises(ObjectDoesNotExist):
        Quote.objects.get(id=quote.id)



import pytest
from django.contrib.auth import get_user_model
from django_quotes.models import (
    GroupStats,
    Quote,
    QuoteStats,
    Source,
    SourceGroup,
    SourceStats,
)
from django_quotes.signals import quote_random_retrieved

from django_markov.models import MarkovTextModel, sentence_generated

User = get_user_model()

pytestmark = pytest.mark.django_db(transaction=True)


def test_sourcegroup_description_render(user: User) -> None:
    """
    Test that a description on source group triggers markdown generation.
    :param user: Logged in user.
    :return:
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    assert group.description is None and group.description == group.description_rendered
    group.description = "A **dark** time for all."
    group.save()
    assert group.description_rendered == "<p>A <strong>dark</strong> time for all.</p>"
    group.description = None
    group.save()
    assert group.description_rendered is None


def test_source_description_render(user: User) -> None:
    """
    Test that a description on a source triggers markdown version.
    :param user:
    :return:
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    source = Source.objects.create(name="Curious George", group=group, owner=user)
    assert source.description is None and source.description_rendered == source.description
    source.description = "A **dark** time for all."
    source.save()
    assert source.description_rendered == "<p>A <strong>dark</strong> time for all.</p>"
    source.description = None
    source.save()
    assert source.description_rendered is None


def test_quote_rendering(user: User) -> None:
    """
    Test that quotes get properly formatted as markdown.
    :param user: The logged in user
    :return:
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    source = Source.objects.create(name="Curious George", group=group, owner=user)
    # Quotes can never be none so we just have to verify that the rendered version is getting created.
    quote = Quote.objects.create(source=source, owner=user, quote="A **dark** time for all.")
    assert quote.quote_rendered == "<p>A <strong>dark</strong> time for all.</p>"


def test_source_creation_allow_creates_markov_model_object(user: User) -> None:
    """
    Test that creating a source also creates it's related markov model with initially empty data.
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    source = Source.objects.create(name="Curious George", group=group, owner=user)
    assert source.text_model


def test_group_creation_creates_markov_model(user: User) -> None:
    """
    Ensure that creating a source group also creates a GroupMarkov instance.
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    assert group.text_model


def test_source_group_creation_generates_stats_object(user: User) -> None:
    """
    Test that the creation of either a source group or source creates its related
    stats object.
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    assert GroupStats.objects.get(group=group)
    source = Source.objects.create(name="Curious George", group=group, owner=user)
    assert SourceStats.objects.get(source=source)
    quote = Quote.objects.create(quote="I'm all a twitter.", source=source, owner=user)
    assert QuoteStats.objects.get(quote=quote)


@pytest.fixture
def statable_source(user):
    """
    Create a source suitable for stat collection.
    :param user: A user who owns the record.
    :return: Source object
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    source = Source.objects.create(name="Curious George", group=group, owner=user)
    Quote.objects.create(quote="Silly little monkey.", source=source, owner=user)
    Quote.objects.create(
        quote="I need bananas or else I get the shakes.",
        source=source,
        owner=user,
    )
    Quote.objects.create(quote="I'm going to take your thumbs first.", source=source, owner=user)
    yield source
    group.delete()


def test_quote_retrieve_stat_signal(statable_source):
    """
    Test that stat are updated correctly when the signal is fired.
    :param statable_source: An instance of Source with stat objects attached
    """
    char_quotes_requested = statable_source.stats.quotes_requested
    group_quotes_requested = statable_source.group.stats.quotes_requested
    quote = statable_source.quote_set.all()[0]
    quote_usage = quote.stats.times_used
    quote_random_retrieved.send(Source, instance=statable_source, quote_retrieved=quote)
    statable_source.refresh_from_db()
    quote.refresh_from_db()
    assert char_quotes_requested < statable_source.stats.quotes_requested
    assert group_quotes_requested < statable_source.group.stats.quotes_requested
    assert quote_usage < quote.stats.times_used


def test_markov_stat_signal(statable_source):
    char_quotes_generated = statable_source.stats.quotes_generated
    group_quotes_generated = statable_source.group.stats.quotes_generated
    sentence_generated.send(
        MarkovTextModel, instance=statable_source.text_model, char_limit=280, sentence="We all go mad sometimes."
    )
    statable_source.refresh_from_db()
    assert char_quotes_generated < statable_source.stats.quotes_generated
    assert group_quotes_generated < statable_source.group.stats.quotes_generated


# def test_quote_create_edit_markov_generation_signal(property_group):
#     source = property_group.source_set.filter(allow_markov=True)[0]
#     cmodel_lastmodify = SourceMarkovModel.objects.get(source=source).modified
#     gmodel_lastmodify = GroupMarkovModel.objects.get(group=property_group).modified
#     q = Quote.objects.create(
#         quote="I am a new quote, full of exciting things to think about.",
#         source=source,
#         owner=property_group.owner,
#         citation="Some episode",
#     )
#     assert (
#         cmodel_lastmodify
#         < SourceMarkovModel.objects.get(source=source).modified
#     )
#     assert (
#         gmodel_lastmodify < GroupMarkovModel.objects.get(group=property_group).modified
#     )
#     cmodel_lastmodify = SourceMarkovModel.objects.get(source=source).modified
#     gmodel_lastmodify = GroupMarkovModel.objects.get(group=property_group).modified
#     q.quote = "Let's change things up a bit with a new quote."
#     q.save()
#     assert (
#         cmodel_lastmodify
#         < SourceMarkovModel.objects.get(source=source).modified
#     )
#     assert (
#         gmodel_lastmodify < GroupMarkovModel.objects.get(group=property_group).modified
#     )
#
#
def test_source_set_to_allow_markov_regenerates_models(property_group):
    source = property_group.source_set.filter(allow_markov=False)[0]
    cmodel_lastmodify = source.text_model.modified
    gmodel_lastmodify = property_group.text_model.modified
    source.allow_markov = True
    source.save()
    assert cmodel_lastmodify < MarkovTextModel.objects.get(pk=source.text_model.pk).modified
    assert gmodel_lastmodify < MarkovTextModel.objects.get(pk=property_group.text_model.pk).modified



from typing import TYPE_CHECKING

from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView as GenericDetail
from django.views.generic import ListView as GenericList
from django.views.generic.edit import CreateView as GenericCreate
from django.views.generic.edit import DeleteView as GenericDelete
from django.views.generic.edit import UpdateView as GenericUpdate
from rules.contrib.views import PermissionRequiredMixin

from django_quotes.models import Quote, Source, SourceGroup

# Create your views here.


# SOURCE GROUPS


class SourceGroupListView(LoginRequiredMixin, GenericList):
    """
    Displays Source Groups owned by the user.
    TODO: For now, only user owned groups, we won't bother with public options.

    Available at /groups/
    """

    model = SourceGroup
    context_object_name = "groups"
    template_name = "quotes/group_list.html"
    paginate_by = 15
    allow_empty = True

    def get_queryset(self):
        return SourceGroup.objects.filter(owner=self.request.user)  # type: ignore


class SourceGroupDetailView(LoginRequiredMixin, PermissionRequiredMixin, GenericDetail):
    """
    Displays details for a source group.

    Available at /groups/[group_slug]/
    """

    model = SourceGroup
    context_object_name = "group"
    template_name = "quotes/group_detail.html"
    permission_required = "django_quotes.read_sourcegroup"
    slug_url_kwarg = "group"
    prefetch_related = ["source_set"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["source_sample"] = Source.objects.filter(group=context["group"])[:5]
        return context


class SourceGroupUpdateView(LoginRequiredMixin, PermissionRequiredMixin, GenericUpdate):
    """
    Update an existing source group

    Available at /groups/[group_slug]/edit/
    """

    if TYPE_CHECKING:
        object: SourceGroup

    model = SourceGroup
    context_object_name = "group"
    template_name = "quotes/group_update.html"
    permission_required = "django_quotes.edit_sourcegroup"
    fields = ["name", "description", "public"]
    slug_url_kwarg = "group"

    def get_success_url(self):
        messages.success(self.request, _(f"Successfully updated group {self.object.name}!"))
        return reverse_lazy("quotes:group_detail", kwargs={"group": self.object.slug})


class SourceGroupDeleteView(LoginRequiredMixin, PermissionRequiredMixin, GenericDelete):  # type: ignore
    """
    Delete and existing source group.

    Available at /groups/[group_slug]/delete/
    """

    model = SourceGroup
    context_object_name = "group"
    template_name = "quotes/group_delete.html"
    permission_required = "django_quotes.delete_sourcegroup"
    slug_url_kwarg = "group"
    success_url = reverse_lazy("quotes:group_list")


class SourceGroupCreateView(LoginRequiredMixin, GenericCreate):
    """
    Create a new source group.

    Available at /groups/create/
    """

    model = SourceGroup
    template_name = "quotes/group_create.html"
    fields = ["name", "description", "public"]

    def form_valid(self, form):
        form.instance.owner = self.request.user
        obj = form.save()
        messages.success(self.request, _(f"Successfully created group {obj.name}"))
        return HttpResponseRedirect(redirect_to=reverse_lazy("quotes:group_detail", kwargs={"group": obj.slug}))


# SOURCES


class SourceCreateView(LoginRequiredMixin, PermissionRequiredMixin, GenericCreate):
    """
    Create a new source and add them to a source group.

    Available at /groups/[group_slug]/sources/add/
    """

    model = Source
    template_name = "quotes/source_create.html"
    fields = ["name", "description", "allow_markov", "public"]
    permission_required = "django_quotes.edit_sourcegroup"
    group = None

    def dispatch(self, request, *args, **kwargs):
        group_slug = kwargs.pop("group")
        self.group = get_object_or_404(SourceGroup, slug=group_slug)
        return super().dispatch(request, *args, **kwargs)

    def get_permission_object(self):
        return self.group

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["group"] = self.group
        return context

    def form_valid(self, form):
        form.instance.owner = self.request.user
        form.instance.group = self.group
        obj = form.save()
        messages.success(self.request, _(f"Successfully created source {obj.name}!"))
        return HttpResponseRedirect(redirect_to=reverse_lazy("quotes:source_detail", kwargs={"source": obj.slug}))


class SourceDetailView(LoginRequiredMixin, PermissionRequiredMixin, GenericDetail):
    """
    Shows information about the specified source.

    Available at /groups/[group_slug]/sources/[source_slug]/
    """

    model = Source
    slug_url_kwarg = "source"
    slug_field = "slug"
    template_name = "quotes/source_detail.html"
    context_object_name = "source"
    permission_required = "django_quotes.read_source"
    prefetch_related = "quote_set"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["samp_quotes"] = context["source"].quote_set.all()[:5]
        return context


class SourceUpdateView(LoginRequiredMixin, PermissionRequiredMixin, GenericUpdate):
    """
    Allows editing and updating of a Source.

    Available at /groups/[group_slug]/sources/[source_slug]/edit/
    """

    model = Source
    slug_url_kwarg = "source"
    slug_field = "slug"
    template_name = "quotes/source_update.html"
    permission_required = "django_quotes.edit_source"
    fields = ["name", "description", "public", "allow_markov"]

    def get_success_url(self):
        messages.success(self.request, "Successfully updated source!")
        return reverse_lazy("quotes:source_detail", kwargs={"source": self.kwargs["source"]})


class SourceDeleteView(LoginRequiredMixin, PermissionRequiredMixin, GenericDelete):  # type: ignore
    """
    Used to delete a given Source

    Available at /groups/[group_slug]/sources/[source_slug]/delete/
    """

    model = Source
    slug_field = "slug"
    slug_url_kwarg = "source"
    template_name = "quotes/source_delete.html"
    permission_required = "django_quotes.delete_source"
    context_object_name = "source"

    def dispatch(self, request, *args, **kwargs):
        source_slug = kwargs.get("source")
        source = get_object_or_404(Source, slug=source_slug)  # type: ignore
        self.group = source.group
        return super().dispatch(request, *args, **kwargs)

    def get_success_url(self):
        return reverse_lazy("quotes:source_list", kwargs={"group": self.group.slug})


class SourceListView(LoginRequiredMixin, PermissionRequiredMixin, GenericList):
    """
    Display a list of sources for a given group.

    Available at /groups/[group_slug]/sources/
    """

    model = Source
    template_name = "quotes/source_list.html"
    permission_required = "django_quotes.read_sourcegroup"
    context_object_name = "sources"
    paginate_by = 15
    allow_empty = True
    group = None

    def dispatch(self, request, *args, **kwargs):
        group_slug = kwargs.pop("group")
        self.group = get_object_or_404(SourceGroup, slug=group_slug)
        return super().dispatch(request, *args, **kwargs)

    def get_permission_object(self):
        return self.group

    def get_queryset(self):
        return (
            Source.objects.filter(group=self.group)
            .select_related("group", "owner")
            .prefetch_related("quote_set")
            .order_by("name")
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["group"] = self.group
        return context


# QUOTES


class QuoteListView(LoginRequiredMixin, PermissionRequiredMixin, GenericList):
    """
    View for viewing quotes from a specific source.

    Available at /groups/[group_slug]/sources/[source_slug]/quotes/
    """

    model = Quote
    context_object_name = "quotes"
    template_name = "quotes/quote_list.html"
    permission_required = "django_quotes.read_source"
    paginate_by = 15
    allow_empty = True
    source = None

    def dispatch(self, request, *args, **kwargs):
        source_slug = kwargs.pop("source")
        self.source = get_object_or_404(Source, slug=source_slug)
        return super().dispatch(request, *args, **kwargs)

    def get_queryset(self):
        return Quote.objects.filter(source=self.source).select_related("source", "source__group").order_by("-created")

    def get_permission_object(self):
        return self.source

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["source"] = self.source
        return context


class QuoteCreateView(LoginRequiredMixin, PermissionRequiredMixin, GenericCreate):
    """
    View for adding a quote to a source.

    Available at /groups/[group_slug]/sources/[source_slug]/quotes/add/
    """

    model = Quote
    template_name = "quotes/quote_create.html"
    permission_required = "django_quotes.edit_source"
    fields = ["quote", "citation", "citation_url", "pub_date"]

    def dispatch(self, request, *args, **kwargs):
        source_slug = kwargs.pop("source")
        self.source = get_object_or_404(Source, slug=source_slug)  # type: ignore
        return super().dispatch(request, *args, **kwargs)

    def get_permission_object(self):
        return self.source

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["source"] = self.source
        return context

    def form_valid(self, form):
        # Check to see if this exact quote is already here. This is imperfect, but better than nothing.
        if Quote.objects.filter(source=self.source, quote=form.instance.quote).exists():
            messages.error(
                self.request,
                _(f"This quote for source {self.source.name} already exists."),
            )
            return self.form_invalid(form)
        form.instance.source = self.source
        form.instance.owner = self.request.user
        form.save()
        messages.success(self.request, _("Successfully added quote!"))
        return HttpResponseRedirect(redirect_to=reverse_lazy("quotes:quote_list", kwargs={"source": self.source.slug}))


class QuoteUpdateView(LoginRequiredMixin, PermissionRequiredMixin, GenericUpdate):
    """
    View for updating a quote.

    Available at /groups/[group_slug]/sources/[source_slug]/quotes/[quote_id]/edit/
    """

    model = Quote
    permission_required = "django_quotes.edit_quote"
    template_name = "quotes/quote_update.html"
    fields = ["quote", "citation", "citation_url", "pub_date"]
    pk_url_kwarg = "quote"

    def form_valid(self, form):
        obj = form.save()
        messages.success(self.request, _("Successfully updated quote!"))
        return HttpResponseRedirect(redirect_to=reverse_lazy("quotes:quote_detail", kwargs={"quote": obj.id}))


class QuoteDetailView(LoginRequiredMixin, PermissionRequiredMixin, GenericDetail):
    """
    A quote detail view for use in things like previewing render or viewing statistics.

    Available at /groups/[group_slug]/sources/[source_slug]/quotes/[quote_id]/
    """

    model = Quote
    pk_url_kwarg = "quote"
    context_object_name = "quote"
    permission_required = "django_quotes.read_quote"
    template_name = "quotes/quote_detail.html"
    select_related = ["source", "source__group", "owner"]


class QuoteDeleteView(LoginRequiredMixin, PermissionRequiredMixin, GenericDelete):  # type: ignore
    """
    View to delete a quote.

    Available at /groups/[group_slug]/sources/[source_slug]/quotes/[quote_id]/delete/
    """

    model = Quote
    pk_url_kwarg = "quote"
    context_object_name = "quote"
    permission_required = "django_quotes.delete_quote"
    template_name = "quotes/quote_delete.html"

    def get_object(self, *args, **kwargs):
        object = super().get_object(*args, **kwargs)
        self.source = object.source  # type: ignore
        return object

    def get_success_url(self):
        return reverse_lazy("quotes:quote_list", kwargs={"source": self.source.slug})



from django.contrib import admin

from django_quotes.models import (
    GroupStats,
    Quote,
    QuoteStats,
    Source,
    SourceGroup,
    SourceStats,
)

# Register your models here.


class SourceGroupAdmin(admin.ModelAdmin):
    """Model admin for SourceGroup."""

    pass


class SourceAdmin(admin.ModelAdmin):
    """Model admin for Source."""

    pass


class QuoteAdmin(admin.ModelAdmin):
    """Model admin for Quote."""

    pass


class GroupStatsAdmin(admin.ModelAdmin):
    """Model admin for GroupStats"""

    pass


class SourceStatAdmin(admin.ModelAdmin):
    """Model admin for SourceStats"""

    pass


class QuoteStatAdmin(admin.ModelAdmin):
    """Model admin for QuoteStats"""

    pass


admin.site.register(SourceGroup, SourceGroupAdmin)
admin.site.register(Source, SourceAdmin)
admin.site.register(Quote, QuoteAdmin)
admin.site.register(GroupStats, GroupStatsAdmin)
admin.site.register(SourceStats, SourceStatAdmin)
admin.site.register(QuoteStats, QuoteStatAdmin)



from __future__ import annotations

import re

import pytest
from django.contrib.auth import get_user_model
from django_quotes.models import SourceGroup
from django_quotes.utils import generate_unique_slug_for_model

pytestmark = pytest.mark.django_db(transaction=True)

User = get_user_model()


def test_unique_generate_unique_slug(user: User):
    sg = SourceGroup.objects.create(name="Explorers Wanted", slug="ew", owner=user)  # Would have a slug of `ew`
    result = generate_unique_slug_for_model(SourceGroup, "EW")
    assert result != sg.slug


@pytest.mark.parametrize(
    "max_length,int_like,slug_truncated",
    [
        (17, True, True),
        (18, False, False),
    ],
)
def test_unique_slug_conforms_to_specified_length(
    user: User, max_length: int | None, int_like: bool, slug_truncated: bool
):
    ends_in_number = re.compile(r"\d+$")
    podcast_title: str = "I'm a podcast with a following"
    control_sample_slug: str = "i-m-a-podcast-wit"
    SourceGroup.objects.create(name=podcast_title, slug=control_sample_slug, owner=user)
    result = generate_unique_slug_for_model(SourceGroup, podcast_title, max_length_override=max_length)
    assert len(result) <= max_length
    if int_like:
        assert ends_in_number.search(result)
    if slug_truncated:
        assert control_sample_slug not in result



from django.conf import settings
from django_quotes.api.views import SourceGroupViewSet, SourceViewSet
from rest_framework.routers import DefaultRouter, SimpleRouter

if settings.DEBUG:
    router = DefaultRouter()
else:
    router = SimpleRouter()

router.register("groups", SourceGroupViewSet, basename="group")
router.register("sources", SourceViewSet, basename="source")


app_name = "api"
urlpatterns = router.urls



# Generated by Django 5.0.4 on 2024-04-03 20:44

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_quotes", "0001_squashed_0005_alter_groupmarkovmodel_created_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="source",
            name="description_rendered",
            field=models.TextField(
                blank=True, help_text="Automatically generated from description on save.", null=True
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-04 19:52

from django.db import migrations


def move_data_from_quotes_django_markov(apps, schema_editor):  # no cov
    Source = apps.get_model("django_quotes", "Source")
    SourceGroup = apps.get_model("django_quotes", "SourceGroup")
    SourceMarkovModel = apps.get_model("django_quotes", "SourceMarkovModel")
    GroupMarkovModel = apps.get_model("django_quotes", "GroupMarkovModel")
    MarkovTextModel = apps.get_model("django_markov", "MarkovTextModel")
    for group in SourceGroup.objects.all():
        g_m = MarkovTextModel.objects.create()
        group.text_model = g_m
        group.save()
        group_model = GroupMarkovModel.objects.get(group=group)
        g_m.data = group_model.data
        g_m.save()

    for source in Source.objects.all():
        s_m = MarkovTextModel.objects.create()
        source.text_model = s_m
        source.save()
        source_model = SourceMarkovModel.objects.get(source=source)
        s_m.data = source_model.data
        s_m.save()


def reverse_data_migration(apps, schema_editor):  # no cov
    Source = apps.get_model("django_quotes", "Source")
    SourceGroup = apps.get_model("django_quotes", "SourceGroup")
    MarkovTextModel = apps.get_model("django_markov", "MarkovTextModel")
    for source in Source.objects.all():
        if source.text_model:
            text_model = MarkovTextModel.objects.get(pk=source.text_model.id)
            source.text_model = None
            source.save()
            text_model.delete()
    for group in SourceGroup.objects.all():
        if group.text_model:
            text_model = MarkovTextModel.objects.get(pk=group.text_model.id)
            group.text_model = None
            group.save()
            text_model.delete()


class Migration(migrations.Migration):

    dependencies = [
        ("django_quotes", "0007_source_text_model_sourcegroup_text_model"),
        ("django_markov", "0005_alter_markovtextmodel_data"),
    ]

    operations = [migrations.RunPython(move_data_from_quotes_django_markov, reverse_code=reverse_data_migration)]



"""Access control rules."""

import rules


@rules.predicate
def is_owner(user, obj):
    """Check if the user is the owner of the object."""
    return obj.owner == user


@rules.predicate
def is_public(user, obj):
    """Check if the object is public."""
    return obj.public


@rules.predicate
def allows_submissions(user, obj):
    """Check if the object allows submissions."""
    return obj.allow_submissions


is_owner_or_public = is_public | is_owner  # type: ignore


@rules.predicate
def is_group_owner(user, obj):
    """Check if the object's group is owned by the user."""
    return user == obj.group.owner


@rules.predicate
def is_source_owner(user, obj):
    """Check that the quote source is owned by the user."""
    return user == obj.source.owner


is_group_owner_and_authenticated = rules.is_authenticated & is_group_owner  # type: ignore



import pytest
from django.test import RequestFactory
from django.urls import reverse
from django_quotes.api.serializers import SourceSerializer
from django_quotes.api.views import SourceGroupViewSet
from django_quotes.models import Source, SourceGroup
from rest_framework import status
from rest_framework.test import APIClient

from tests.factories.users import UserFactory

pytestmark = pytest.mark.django_db(transaction=True)


@pytest.fixture
def apiclient():
    return APIClient()


class TestGroupViewSet:
    def test_list_groups(self, property_group: SourceGroup, rf: RequestFactory) -> None:
        view = SourceGroupViewSet()
        request = rf.get("/ignorethisurl/")
        request.user = property_group.owner
        view.request = request
        assert property_group in view.get_queryset()

    def test_retrieve_group(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(
            reverse("api:group-detail", kwargs={"group": property_group.slug}),
            format="json",
        )
        assert response.status_code == status.HTTP_200_OK

    def test_empty_random_quote(self, apiclient):
        user = UserFactory()
        apiclient.force_authenticate(user=user)
        group = SourceGroup.objects.create(name="Nothing here", owner=user)
        response = apiclient.get(
            reverse("api:group-get-random-quote", kwargs={"group": group.slug}),
            format="json",
        )
        assert response.status_code == status.HTTP_404_NOT_FOUND

    def test_random_quote(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(
            reverse(
                "api:group-get-random-quote",
                kwargs={"group": property_group.slug},
            )
        )
        assert response.status_code == status.HTTP_200_OK

    def test_markov_group(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(reverse("api:group-generate-sentence", kwargs={"group": property_group.slug}))
        assert response.status_code in [status.HTTP_200_OK, status.HTTP_204_NO_CONTENT]

    def test_disallowed_markov_group(self, apiclient):
        user = UserFactory()
        group = SourceGroup.objects.create(name="Nothing here", owner=user)
        Source.objects.create(name="No Fun", group=group, owner=user, allow_markov=False)
        apiclient.force_authenticate(user=user)
        response = apiclient.get(reverse("api:group-generate-sentence", kwargs={"group": group.slug}))
        assert response.status_code == status.HTTP_403_FORBIDDEN

    def test_empty_markov_group(self, apiclient):
        user = UserFactory()
        group = SourceGroup.objects.create(name="Nothing here", owner=user)
        Source.objects.create(name="No Fun", group=group, owner=user, allow_markov=True)
        apiclient.force_authenticate(user=user)
        response = apiclient.get(reverse("api:group-generate-sentence", kwargs={"group": group.slug}))
        assert response.status_code == status.HTTP_204_NO_CONTENT


class TestSourceViewSet:
    def test_list_sources(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(reverse("api:source-list"))
        assert response.status_code == status.HTTP_200_OK

    def test_list_sources_by_group(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        extra_group = SourceGroup.objects.create(name="I'm extra", owner=property_group.owner)
        Source.objects.create(name="Bobble the Odd", group=extra_group, owner=property_group.owner)
        response = apiclient.get(reverse("api:source-list") + f"?group={property_group.slug}")
        assert response.status_code == status.HTTP_200_OK
        assert len(response.data) == property_group.source_set.count()

    def test_filter_by_nonexistent_group(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(reverse("api:source-list") + "?group=sayonara")
        assert len(response.data) == 0
        assert response.status_code == status.HTTP_200_OK

    def test_retrieve_source(self, apiclient, property_group):
        char_to_retrieve = property_group.source_set.all()[0]
        serializer = SourceSerializer(char_to_retrieve)
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(reverse("api:source-detail", kwargs={"source": char_to_retrieve.slug}))
        assert response.status_code == status.HTTP_200_OK
        assert response.data == serializer.data

    def test_source_random_quote(self, apiclient, property_group):
        char_to_retrieve = property_group.source_set.all()[0]
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(
            reverse(
                "api:source-get-random-quote",
                kwargs={"source": char_to_retrieve.slug},
            )
        )
        assert response.status_code == status.HTTP_200_OK

    def test_random_quote_empty_source(self, apiclient, property_group):
        apiclient.force_authenticate(user=property_group.owner)
        source = Source.objects.create(name="Bobble the Elder", group=property_group, owner=property_group.owner)
        response = apiclient.get(reverse("api:source-get-random-quote", kwargs={"source": source.slug}))
        assert response.status_code == status.HTTP_404_NOT_FOUND

    def test_source_generate_sentence(self, apiclient, property_group):
        char_to_retrieve = property_group.source_set.filter(allow_markov=True)[0]
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(
            reverse(
                "api:source-generate-sentence",
                kwargs={"source": char_to_retrieve.slug},
            )
        )
        assert response.status_code in [status.HTTP_200_OK, status.HTTP_204_NO_CONTENT]
        if response.status_code == status.HTTP_200_OK:
            assert response.data["sentence"] is not None

    def test_disallowed_source_generate_sentence(self, apiclient, property_group):
        char_to_retrieve = property_group.source_set.filter(allow_markov=False)[0]
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(
            reverse(
                "api:source-generate-sentence",
                kwargs={"source": char_to_retrieve.slug},
            )
        )
        assert response.status_code == status.HTTP_403_FORBIDDEN

    def test_empty_source_generate_sentence(self, apiclient, property_group):
        char_to_retrieve = Source.objects.create(
            name="I say nothing",
            group=property_group,
            owner=property_group.owner,
            allow_markov=True,
        )
        apiclient.force_authenticate(user=property_group.owner)
        response = apiclient.get(
            reverse(
                "api:source-generate-sentence",
                kwargs={"source": char_to_retrieve.slug},
            )
        )
        assert response.status_code == status.HTTP_204_NO_CONTENT



from io import StringIO

import pytest
from django.core.management import call_command
from django_quotes.models import SourceGroup

from django_markov.models import MarkovTextModel

pytestmark = pytest.mark.django_db(transaction=True)


def test_markov_command(property_group):
    pgmm = property_group.text_model
    source = property_group.source_set.filter(allow_markov=True)[0]
    group_modify = pgmm.modified
    cmm = source.text_model
    char_modify = cmm.modified
    nochangegroup = SourceGroup.objects.create(
        name="So Alone", owner=property_group.owner, text_model=MarkovTextModel.objects.create()
    )
    nochange_modify = nochangegroup.text_model.modified
    out = StringIO()
    call_command("makemarkov", stdout=out, stderr=StringIO())
    assert nochange_modify == MarkovTextModel.objects.get(pk=nochangegroup.text_model.pk).modified
    pgmm.refresh_from_db()
    cmm.refresh_from_db()
    assert group_modify < pgmm.modified
    assert char_modify < cmm.modified
    assert pgmm.data is not None
    assert cmm.data is not None



# Generated by Django 5.0.4 on 2024-04-04 19:51

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_markov", "0005_alter_markovtextmodel_data"),
        ("django_quotes", "0006_alter_source_description_rendered"),
    ]

    operations = [
        migrations.AddField(
            model_name="source",
            name="text_model",
            field=models.OneToOneField(
                help_text="The text model for this character.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="django_markov.markovtextmodel",
            ),
        ),
        migrations.AddField(
            model_name="sourcegroup",
            name="text_model",
            field=models.OneToOneField(
                help_text="The markov model for this group.",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="django_markov.markovtextmodel",
            ),
        ),
    ]



from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model
from loguru import logger
from slugify import slugify


def generate_unique_slug_for_model(
    model_class: type[Model],
    text: str,
    slug_field: str | None = "slug",
    max_length_override: int | None = None,
) -> str:
    """
    Generate a unique slug for the given model.

    Args:
        model_class (Model): A class based upon ``django.db.models.Model``
        text (str): Text to convert to a slug.
        slug_field (str | None): The name of the slug field of the model.
        max_length_override (int | None): Maximum number of characters to use if not the same as what's defined in the slug field.
    Returns:
         (str): The generated slug.
    """
    unique_found: bool = False
    has_next: bool = False
    next_val: int = 0
    if not max_length_override:
        logger.debug("Setting max_length of slug from field definition.")
        max_length: int = model_class._meta.get_field(slug_field).max_length  # type: ignore
    else:
        logger.debug("User override value for max length of slug.")
        max_length = max_length_override
    slug = slugify(text, max_length=max_length)
    logger.debug(f"Base slug is set to '{slug}'.")
    while not unique_found:
        logger.debug(f"Testing uniqueness of slug '{slug}'...")
        try:
            model_class.objects.get(**{str(slug_field): slug})
        except ObjectDoesNotExist:
            logger.debug("Slug is unique!")
            unique_found = True
        if not unique_found:
            logger.debug("Slug is not unique yet.")
            next_val += 1
            if has_next:
                slug = slug[len(slug) - (len(str(next_val - 1)) - 1) :]
            if len(slug) >= max_length:
                slug = slug[: max_length - (len(str(next_val)) + 1)]
            slug = slug + f"-{next_val}"
            has_next = True
    return slug



from __future__ import annotations

import random
from typing import TYPE_CHECKING, Any

import rules
from asgiref.sync import async_to_sync

if TYPE_CHECKING:
    from django.db.models.manager import RelatedManager
from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from loguru import logger
from rules.contrib.models import RulesModelBase, RulesModelMixin

from django_markov.models import MarkovTextModel
from django_quotes.rules import (  # is_character_owner,; is_group_owner_and_authenticated,
    is_owner,
    is_owner_or_public,
)
from django_quotes.signals import quote_random_retrieved
from django_quotes.utils import generate_unique_slug_for_model

MAX_QUOTES_FOR_RANDOM_SET = 50
MAX_QUOTES_FOR_RANDOM_GROUP_SET = 50

if hasattr(settings, "MAX_QUOTES_FOR_RANDOM_SET"):  # pragma: nocover
    MAX_QUOTES_FOR_RANDOM_SET = settings.MAX_QUOTES_FOR_RANDOM_SET

if hasattr(settings, "MAX_QUOTES_FOR_RANDOM_GROUP_SET"):  # pragma: nocover
    MAX_QUOTES_FOR_RANDOM_GROUP_SET = settings.MAX_QUOTES_FOR_RANDOM_GROUP_SET


class AbstractOwnerModel(models.Model):
    """
    Abstract model for representing an entity owned by a user with toggles for either allowing submissions for it
    and public access. Defaults to completely private by default.

    Attributes:
        public (bool): is this object public to any authenticated user? Default: False
        allow_submissions (bool): allow other users to submit child objects? Default: False. Not implemented yet.
        owner (User): The user that created this object.
    """

    public = models.BooleanField(
        default=False,
        help_text=_("Is this a public source available for any user to view?"),
    )
    allow_submissions = models.BooleanField(default=False, help_text=_("Allow submissions from other users?"))
    owner = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)

    class Meta:
        abstract = True


class TimeStampedModel(models.Model):
    """Automatically adds a created and modified field."""

    created = models.DateTimeField(auto_now_add=True, help_text=_("Created timestamp"))
    modified = models.DateTimeField(auto_now=True, help_text=_("Last modified time"))

    class Meta:
        abstract = True


# Create your models here.
class SourceGroup(AbstractOwnerModel, RulesModelMixin, TimeStampedModel, metaclass=RulesModelBase):
    """
    An abstract group or source for a given set of quotes. Multiple sources, or Source objects, can belong to
    the same group. For example, a novel or series if you plan to quote the characters within individually.

    Attributes:
        id (int): Database primary key for the object.
        name (str): Human readable string to name the group. This will be converted to a slug prefix.
        description (str): A description of the group for convenience. Markdown can be used here for styling.
        description_rendered (str): The HTML representation of the description string. Generated automatically.
        owner (User): The user that created the group and therefore owns it.
        public (bool): Is this group public or private. Defaults to False.
        allow_submissions (bool): Allow other users to submit characters to this. Not yet implemented.
        slug (str): A unique slug to represent this group. Generated automatically from name.
        created (datetime): When this object was first created. Auto-generated.
        modified (datetime): Last time this object was modified. Auto-generated.

    """

    if TYPE_CHECKING:
        source_set: RelatedManager[Source]

    name = models.CharField(
        _("Source Name"),
        max_length=50,
        help_text=_("A source for individuals making the quotes. Use as an abstract grouping."),
        db_index=True,
    )
    description = models.TextField(
        help_text=_("Description for the source. You can style using Markdown."),
        null=True,
        blank=True,
    )
    description_rendered = models.TextField(
        help_text=_("Automatically generated from description"), null=True, blank=True
    )
    slug = models.SlugField(
        unique=True,
        max_length=70,
        editable=False,
        blank=True,
        help_text=_("Unique slug for this group."),
    )
    text_model = models.OneToOneField(
        MarkovTextModel,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text=_("The markov model for this group."),
    )

    class Meta:
        rules_permissions = {
            "add": rules.is_authenticated,
            "read": is_owner_or_public,
            "edit": is_owner,
            "delete": is_owner,
        }
        ordering = ["name"]

    def __str__(self):  # pragma: nocover
        return self.name

    def save(self, *args, **kwargs):
        """Save and create slug if missing."""
        if not self.slug:  # Once this slug is set, it does not change except through devil pacts
            logger.debug("Group is being saved and a slug was provided.")
            self.slug = generate_unique_slug_for_model(model_class=type(self), text=self.name)
        super().save(*args, **kwargs)

    def refresh_from_db(self, *args, **kwargs):
        """
        Also reset cached properties.
        """
        super().refresh_from_db(*args, **kwargs)
        cached_properties = ["total_sources", "markov_sources", "total_quotes"]
        for prop in cached_properties:
            try:
                del self.__dict__[prop]
            except KeyError:  # pragma: nocover
                pass

    @cached_property
    def total_sources(self) -> int:
        """Total number of sources for the group."""
        return Source.objects.filter(group=self).count()

    @cached_property
    def markov_sources(self) -> int:
        """Total number of Markov sources for the group."""
        return Source.objects.filter(group=self, allow_markov=True).count()

    @cached_property
    def total_quotes(self) -> int:
        """Total quotes for the group."""
        return Quote.objects.filter(source__in=Source.objects.filter(group=self)).count()

    @cached_property
    def markov_ready(self) -> bool:
        """Checks to see if there are Markov enabled sources and sufficient quotes."""
        if (
            self.markov_sources > 0
            and Quote.objects.filter(source__in=self.source_set.filter(allow_markov=True)).count() > 10  # noqa:PLR2004
        ):
            return True
        return False

    async def aupdate_markov_model(self) -> None:
        """Updates the related MarkovTextModel."""
        markov_sources = self.source_set.filter(allow_markov=True)
        if await markov_sources.aexists():
            quotes = Quote.objects.filter(source__in=markov_sources)
            await self.text_model.aupdate_model_from_corpus(corpus_entries=[quote.quote async for quote in quotes])

    def update_markov_model(self) -> None:
        """Updates the related MarkovTextModel."""
        async_to_sync(self.aupdate_markov_model)()

    def generate_markov_sentence(self, max_characters: int = 280, tries: int = 20) -> str | None:
        """
        Generate a markov sentence based on quotes from markov enabled characters for the group.

        Args:
            max_characters (int): Maximum characters allowed in the resulting sentence.
            tries (int): Maximum number of tries django_markov should use to create the sentence.

        Returns:
            (str | None): The generated sentence or None if no sentence was possible for the number
                of tries.
        """
        if self.markov_ready:
            logger.debug("Group is ready for markov sentences. Checking model...")
            mmodel = self.text_model
            if not mmodel.is_ready:
                logger.debug("Markov model for group is not generated yet! Generating...")
                self.update_markov_model()
                mmodel.refresh_from_db()
            logger.debug("Generating sentence...")
            sentence: str | None = mmodel.generate_sentence(char_limit=max_characters, tries=tries)
            if sentence is not None:
                logger.debug(f"Returning generated sentence: '{sentence}'")
                return sentence
        logger.debug("Group is not ready for markov requests yet!")
        return None

    def get_random_quote(self, max_quotes_to_process: int | None = MAX_QUOTES_FOR_RANDOM_GROUP_SET) -> Any:
        """
        Get a random quote object from any of the characters defined within the group.
        Prioritizes quotes that have been returned less often.

        Args:
            max_quotes_to_process (int | None): Maximum number of quotes to retrieve before
                selecting a random one.

        Returns:
             (Quote | None) Quote object or None if no quotes are found.
        """
        # TODO: Create Q object to filter for null datetimes or less than now.
        quotes = (
            Quote.objects.filter(source__in=self.source_set.all())
            .filter(models.Q(pub_date__isnull=True) | models.Q(pub_date__lte=timezone.now()))
            .select_related("stats")
            .order_by("stats__times_used")[:max_quotes_to_process]
        )
        if quotes.exists():
            quote = random.choice(list(quotes))  # noqa: S311
            quote_random_retrieved.send(type(quote.source), instance=quote.source, quote_retrieved=quote)
            return quote
        return None


class Source(AbstractOwnerModel, RulesModelMixin, TimeStampedModel, metaclass=RulesModelBase):
    """
    An individual source to attribute the quote to in the system, such as a character from a podcast/book, or a specific
    author. A user must be the owner of the related SourceGroup to add or delete a source.

    Attributes:
        id (int): Database primary key for the object.
        name (str): Unique name of a character within a ``CharacterGroup`` for this entity.
        group (SourceGroup): The parent ``SourceGroup``.
        slug (str): Slug made up of a generated version of the character name and the group slug prefix.
        description (str): Description for the character. Markdown can be used for styling.
        description_rendered (str): HTML representation of the description for convenience. Automatically generated.
        allow_markov (bool): Allow markov quotes to be requested from this character? Default False.
        owner (User): The user that created and owns this character.
        public (bool): Is the character public to other users? Defaults to False.
        allow_submissions (bool): Allow other users to submit quotes for this character? Defaults to False.
        created (datetime): When this object was first created. Auto-generated.
        modified (datetime): Last time this object was modified. Auto-generated.

    """

    if TYPE_CHECKING:
        quote_set: RelatedManager[Quote]

    name = models.CharField(max_length=100, help_text=_("Name of the character"))
    slug = models.SlugField(
        max_length=250,
        help_text=_("Global slug of the character, will be auto generated from name and group if not overridden."),
        blank=True,
        editable=False,
        unique=True,
        db_index=True,
    )
    description = models.TextField(
        null=True,
        blank=True,
        help_text=_("Description of this character. You can style this with Markdown."),
    )
    description_rendered = models.TextField(
        null=True, blank=True, help_text=_("Automatically generated from description on save.")
    )
    allow_markov = models.BooleanField(default=False, help_text=_("Allow to be used in markov chains?"))
    group = models.ForeignKey(
        SourceGroup,
        on_delete=models.CASCADE,
        help_text=_("The group this character belongs to."),
    )
    text_model = models.OneToOneField(
        MarkovTextModel,
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text=_("The text model for this character."),
    )

    class Meta:
        rules_permissions = {
            "add": rules.is_authenticated,
            "read": is_owner_or_public,
            "edit": is_owner,
            "delete": is_owner,
        }

    def __str__(self):  # pragma: nocover
        return self.name

    def save(self, *args, **kwargs):
        """Save and create slug, if missing."""
        if not self.slug:
            self.slug = generate_unique_slug_for_model(type(self), text=f"{self.group.slug} {self.name}")
        super().save(*args, **kwargs)

    @property
    def markov_ready(self) -> bool:
        """
        Conducts sanity checks to see if requesting a markov chain is feasible. Markov must be enabled for a character
        and there must be a sufficient corpus to generate a sentence from. Currently set at a minimum of 10 quotes.

        Returns:
            (bool): If ready for markov requests.
        """
        if self.allow_markov and Quote.objects.filter(source=self).count() > 10:  # noqa:PLR2004
            return True
        return False

    async def _amarkov_ready(self) -> bool:
        """Async version of markov ready.

        Returns:
            (bool): If ready for markov
        """
        if self.allow_markov and await Quote.objects.filter(source=self).acount() > 10:  # noqa: PLR2004
            return True
        return False

    async def aupdate_markov_model(self) -> None:
        """
        Process all quotes into the associated model.
        """
        if self.allow_markov:
            await self.text_model.aupdate_model_from_corpus(
                corpus_entries=[quote.quote async for quote in self.quote_set.all()], char_limit=0, store_compiled=False
            )

    def update_markov_model(self) -> None:
        """
        Sync wrapper around `aupdate_markov_model`.
        """
        async_to_sync(self.aupdate_markov_model)()

    def get_markov_sentence(self, max_characters: int | None = 280, tries: int = 20) -> str | None:
        """
        If valid, generate a markov sentence. If not, return None.

        Args:
            max_characters (int | None): Maximum number of characters allowed in
                resulting sentence.
            tries (int): Number of times django_markov may try to generate sentence.

        Returns:
            (str | None): The resulting sentence or None if a sentence could not be formed.
        """
        logger.debug("Checking to see if character is markov ready...")
        if self.markov_ready:
            logger.debug("It IS ready. Fetching markov model.")
            markov_model = self.text_model
            if not markov_model.is_ready:
                logger.debug("No model defined yet, generating...")
                self.update_markov_model()
            logger.debug("Markov text model loaded. Generating sentence.")
            sentence: str | None = markov_model.generate_sentence(char_limit=max_characters, tries=tries)
            if sentence is not None:
                return sentence
        return None

    def get_random_quote(self, max_quotes_to_process: int | None = MAX_QUOTES_FOR_RANDOM_SET) -> Any | None:
        """
        This actually not all that random. It's going to grab the quotes
        ordered ordered by how infrequently they've been returned, and then grab a random one
        in the set. But for our purposes, it's fine. If there aren't any quotes, it will return None.

        Args:
            max_quotes_to_process (int | None): Maximum number of quotes to retrive before
                selecting at random.

        Returns:
            (Quote | None): The quote object or None if no quotes found.
        """
        quotes_to_pick = (
            Quote.objects.filter(source=self)
            .filter(models.Q(pub_date__isnull=True) | models.Q(pub_date__lte=timezone.now()))
            .select_related("stats")
            .order_by("stats__times_used")[:max_quotes_to_process]
        )
        if quotes_to_pick.exists():
            # Select a random index in the result set.
            quote_to_return = random.choice(list(quotes_to_pick))  # noqa: S311
            quote_random_retrieved.send(type(self), instance=self, quote_retrieved=quote_to_return)
            return quote_to_return
        return None


class Quote(AbstractOwnerModel, RulesModelMixin, TimeStampedModel, metaclass=RulesModelBase):
    """
    A quote from a given source. A user must own the related source to add or delete a quote.

    Attributes:
        id (int): Database primary key for the object.
        quote (str): The quote text to use. You can use Markdown for styling. Must be <= 280 characters for tweets
        quote_rendered (str): HTML rendered version of the quote field. Automatically generated.
        citation (str): Optional description of quote source, e.g. episode number or book title.
        citation_url (str): Optional accompanying URL for the citation.
        character (Source): The source of this quote.
        owner (User): The user that created and owns this quote.
        created (datetime): When this object was first created. Auto-generated.
        modified (datetime): Last time this object was modified. Auto-generated.

    """

    if TYPE_CHECKING:
        id: int

    quote = models.CharField(
        max_length=280,  # Keep the base limit to 280 so that quotes are 'tweetable'
        help_text="Plain text representation of quote. You can use Markdown here.",
    )
    quote_rendered = models.TextField(
        null=True,
        blank=True,
        help_text=_("HTML rendered version of quote generated from quote plain text."),
    )
    source = models.ForeignKey(
        Source,
        on_delete=models.CASCADE,
        help_text=_("The source of this quote, i.e. name."),
    )
    citation = models.CharField(
        max_length=250,
        blank=True,
        null=True,
        help_text=_("Where is this quote from? Episode #, book?"),
    )
    citation_url = models.URLField(null=True, blank=True, help_text=_("URL for citation, if applicable."))
    pub_date = models.DateTimeField(
        null=True,
        blank=True,
        help_text=_("When is the earliest time this should appear in random results?"),
    )

    class Meta:
        rules_permissions = {
            # "add": is_character_owner,
            "read": is_owner_or_public,
            "edit": is_owner,
            "delete": is_owner,
        }

    def __str__(self):  # pragma: nocover
        return f"{self.source.name}: {self.quote}"


class QuoteStats(TimeStampedModel):
    """
    A simple object used to track how often an individual quote is used.

    Attributes:
        id (int): The database primary key of this object.
        quote (Quote): The quote this stat relates to.
        times_used (int): The number of times this has been used by an service such as random quote.
        created (datetime): When this was created.
        modified (datetime): When this was last modified.
    """

    quote = models.OneToOneField(
        Quote,
        on_delete=models.CASCADE,
        related_name="stats",
        help_text=_("The Quote the stats related to."),
    )
    times_used = models.PositiveIntegerField(default=0, help_text=_("Times used for random quotes, etc."))

    def __str__(self):  # pragma: nocover
        return f"Stats for Quote {self.quote.id}"


class GroupStats(TimeStampedModel):
    """
    An object for using to track usage stats for ``CharacterGroup``.

    Attributes:
        group (SourceGroup): The group this is collecting stats for.
        quotes_requested (int): The number of times a quote from this object or its children has been requested.
        quotes_generated (int): The number of times a markov quote has been generated for this or it's children.
    """

    group = models.OneToOneField(SourceGroup, related_name="stats", on_delete=models.CASCADE)
    quotes_requested = models.PositiveIntegerField(
        default=0, help_text=_("Number of time child quotes have been requested.")
    )
    quotes_generated = models.PositiveIntegerField(
        default=0,
        help_text=_("Number of times markov generated quotes have been requested."),
    )

    def __str__(self):  # pragma: nocover
        return f"Stats for Group {self.group.name}"


class SourceStats(TimeStampedModel):
    """
    An object for using to track usage stats for ``Character``.

    Attributes:
        source (Source): The source this is collecting stats for.
        quotes_requested (int): The number of times a quote from this object or its children has been requested.
        quotes_generated (int): The number of times a markov quote has been generated for this or it's children.
    """

    source = models.OneToOneField(Source, related_name="stats", on_delete=models.CASCADE)
    quotes_requested = models.PositiveIntegerField(
        default=0, help_text=_("Number of time child quotes have been requested.")
    )
    quotes_generated = models.PositiveIntegerField(
        default=0,
        help_text=_("Number of times markov generated quotes have been requested."),
    )

    def __str__(self):  # pragma: nocover
        return f"Stats for Source {self.source.name}"



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class QuotesConfig(AppConfig):
    """App configuration for django_quotes app.
    Notably loads the receivers so that signals will be processed.
    """

    default_auto_field = "django.db.models.BigAutoField"
    name = "django_quotes"
    verbose_name = _("Quotes")

    def ready(self):
        """Load the receivers."""
        try:
            import django_quotes.receivers  # noqa F401
        except ImportError:  # pragma: nocover
            pass



from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import django_stubs_ext
from dj_easy_log import load_loguru

if TYPE_CHECKING:
    django_stubs_ext.monkeypatch()

ROOT_DIR = Path(__file__).resolve(strict=True).parent.parent
APPS_DIR = ROOT_DIR
ADMIN_URL = "admin/"

# GENERAL
# -----------------------------------------------------------------------------
DEBUG = True
# Local time zone. Choices are
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# though not all of them may be available with every OS.
# In Windows, this must be set to your system time zone.
TIME_ZONE = "UTC"
# https://docs.djangoproject.com/en/dev/ref/settings/#language-code
LANGUAGE_CODE = "en-us"
# https://docs.djangoproject.com/en/dev/ref/settings/#site-id
SITE_ID = 1
# https://docs.djangoproject.com/en/dev/ref/settings/#use-i18n
USE_I18N = True
# https://docs.djangoproject.com/en/dev/ref/settings/#use-l10n
USE_L10N = True
# https://docs.djangoproject.com/en/dev/ref/settings/#use-tz
USE_TZ = True
# https://docs.djangoproject.com/en/dev/ref/settings/#locale-paths
LOCALE_PATHS = [str(ROOT_DIR / "locale")]

# DATABASES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#databases

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "django_quotes.db",
        "ATOMIC_REQUESTS": True,
    }
}
# https://docs.djangoproject.com/en/stable/ref/settings/#std:setting-DEFAULT_AUTO_FIELD
DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# URLS
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#root-urlconf
ROOT_URLCONF = "tests.urls"

# APPS
# ------------------------------------------------------------------------------
DJANGO_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # "django.contrib.humanize", # Handy template tags
    "django.contrib.admin",
    "django.forms",
]
THIRD_PARTY_APPS = [
    "django_markov",
    "crispy_forms",
    "crispy_bootstrap4",
    "rest_framework",
    "rest_framework.authtoken",
    "corsheaders",
    "drf_spectacular",
    "rules.apps.AutodiscoverRulesConfig",
]

LOCAL_APPS = [
    "django_quotes",
    # Your stuff: custom apps go here
]
# https://docs.djangoproject.com/en/dev/ref/settings/#installed-apps
INSTALLED_APPS = DJANGO_APPS + THIRD_PARTY_APPS + LOCAL_APPS

# AUTHENTICATION
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#authentication-backends
AUTHENTICATION_BACKENDS = [
    "rules.permissions.ObjectPermissionBackend",
    "django.contrib.auth.backends.ModelBackend",
]
# PASSWORDS
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#password-hashers
PASSWORD_HASHERS = ["django.contrib.auth.hashers.MD5PasswordHasher"]
# https://docs.djangoproject.com/en/dev/ref/settings/#auth-password-validators
AUTH_PASSWORD_VALIDATORS = [
    {"NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"},
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]
# MIDDLEWARE
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#middleware
MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "corsheaders.middleware.CorsMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.common.BrokenLinkEmailsMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

# STATIC
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#static-root
STATIC_ROOT = str(ROOT_DIR / "django_quotes/staticfiles")
# https://docs.djangoproject.com/en/dev/ref/settings/#static-url
STATIC_URL = "/static/"
# https://docs.djangoproject.com/en/dev/ref/contrib/staticfiles/#std:setting-STATICFILES_DIRS
STATICFILES_DIRS = [str(APPS_DIR / "django_quotes/static")]
# https://docs.djangoproject.com/en/dev/ref/contrib/staticfiles/#staticfiles-finders
STATICFILES_FINDERS = [
    "django.contrib.staticfiles.finders.FileSystemFinder",
    "django.contrib.staticfiles.finders.AppDirectoriesFinder",
]

# MEDIA
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#media-root
MEDIA_ROOT = str(APPS_DIR / "media")
# https://docs.djangoproject.com/en/dev/ref/settings/#media-url
MEDIA_URL = "/media/"

# TEMPLATES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#templates
TEMPLATES = [
    {
        # https://docs.djangoproject.com/en/dev/ref/settings/#std:setting-TEMPLATES-BACKEND
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        # https://docs.djangoproject.com/en/dev/ref/settings/#dirs
        "DIRS": [str(APPS_DIR / "templates")],
        # https://docs.djangoproject.com/en/dev/ref/settings/#app-dirs
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": DEBUG,
            # https://docs.djangoproject.com/en/dev/ref/settings/#template-context-processors
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    }
]
# https://docs.djangoproject.com/en/dev/ref/settings/#form-renderer
FORM_RENDERER = "django.forms.renderers.TemplatesSetting"

CRISPY_ALLOWED_TEMPLATE_PACKS = "bootstrap4"

CRISPY_TEMPLATE_PACK = "bootstrap4"

# FIXTURES
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#fixture-dirs
FIXTURE_DIRS = (str(APPS_DIR / "fixtures"),)

# SECURITY
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#session-cookie-httponly
SESSION_COOKIE_HTTPONLY = True
# https://docs.djangoproject.com/en/dev/ref/settings/#csrf-cookie-httponly
CSRF_COOKIE_HTTPONLY = True
# https://docs.djangoproject.com/en/dev/ref/settings/#secure-browser-xss-filter
SECURE_BROWSER_XSS_FILTER = True
# https://docs.djangoproject.com/en/dev/ref/settings/#x-frame-options
X_FRAME_OPTIONS = "DENY"
# LOGGING
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#logging
# See https://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {"verbose": {"format": "%(levelname)s %(asctime)s %(module)s " "%(process)d %(thread)d %(message)s"}},
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        }
    },
    "root": {"level": "INFO", "handlers": ["console"]},
    "loggers": {
        "rules": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": True,
        },
    },
}

SECRET_KEY = "UKNSQClo0zUtSNlFq7wJYvMcIqMZrZ2MpwAKjDtBEAbvuJosEg1frjvmNX4HmA46"  # Doesn't matter since this is a test.

# https://docs.djangoproject.com/en/dev/ref/settings/#test-runner
TEST_RUNNER = "django.test.runner.DiscoverRunner"


# django-rest-framework
# -------------------------------------------------------------------------------
# django-rest-framework - https://www.django-rest-framework.org/api-guide/settings/
REST_FRAMEWORK = {
    "DEFAULT_AUTHENTICATION_CLASSES": (
        "rest_framework.authentication.SessionAuthentication",
        "rest_framework.authentication.TokenAuthentication",
    ),
    "DEFAULT_PERMISSION_CLASSES": ("rest_framework.permissions.IsAuthenticated",),
    "DEFAULT_SCHEMA_CLASS": "drf_spectacular.openapi.AutoSchema",
}

# django-cors-headers - https://github.com/adamchainz/django-cors-headers#setup
CORS_URLS_REGEX = r"^/api/.*$"

# By Default swagger ui is available only to admin user. You can change permission classs to change that
# See more configuration options at https://drf-spectacular.readthedocs.io/en/latest/settings.html#settings
SPECTACULAR_SETTINGS = {
    "TITLE": "Django Quotes API",
    "DESCRIPTION": "Documentation of API endpoints of Django Quotes",
    "VERSION": "1.0.0",
    "SERVE_PERMISSIONS": ["rest_framework.permissions.IsAdminUser"],
    "SERVERS": [
        {"url": "https://127.0.0.1:8000", "description": "Local Development server"},
    ],
}
REST_FRAMEWORK["TEST_REQUEST_DEFAULT_FORMAT"] = "json"

# EMAIL
# ------------------------------------------------------------------------------
# https://docs.djangoproject.com/en/dev/ref/settings/#email-backend
EMAIL_BACKEND = "django.core.mail.backends.locmem.EmailBackend"

MARKOV_CORPUS_MAX_CHAR_LIMIT = 0


MAX_QUOTES_FOR_RANDOM_SET = 50


MAX_QUOTES_FOR_RANDOM_GROUP_SET = 50


load_loguru(globals())



from __future__ import annotations

from rest_framework.serializers import ModelSerializer

from django_quotes.models import Quote, Source, SourceGroup


class SourceGroupSerializer(ModelSerializer):
    """
    Serializer for SourceGroup.

    Includes the following fields:

    - name (str)
    - slug (str)
    - description (str)
    - description_rendered (str)
    """

    class Meta:
        model = SourceGroup
        fields = ["name", "slug", "description", "description_rendered"]


class SourceSerializer(ModelSerializer):
    """
    Serializer for Source.

    Includes the following fields:

    - name (str)
    - group (SourceGroupSerializer)
    - slug (str)
    - description (str)
    - description_rendered (str)
    """

    group = SourceGroupSerializer()

    class Meta:
        model = Source
        fields = ["name", "group", "slug", "description", "description_rendered"]


class QuoteSerializer(ModelSerializer):
    """
    Serializer for Quote.

    Includes the following fields:

    - quote (str)
    - quote_rendered (str)
    - source (SourceSerializer)
    - citation (str)
    - citation_url (str)
    """

    source = SourceSerializer()  # type: ignore

    class Meta:
        model = Quote
        fields = ["quote", "quote_rendered", "source", "citation", "citation_url"]



from django.conf import settings
from django.conf.urls.static import static
from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, path
from django.views import defaults as default_views
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView
from rest_framework.authtoken.views import obtain_auth_token

urlpatterns = [
    path(settings.ADMIN_URL, admin.site.urls),
    path("app/", include("django_quotes.urls", namespace="quotes")),
    path(
        "400/",
        default_views.bad_request,
        kwargs={"exception": Exception("Bad Request!")},
    ),
    path(
        "403/",
        default_views.permission_denied,
        kwargs={"exception": Exception("Permission Denied")},
    ),
    path(
        "404/",
        default_views.page_not_found,
        kwargs={"exception": Exception("Page not Found")},
    ),
    path("500/", default_views.server_error),
] + static(settings.MEDIA_URL, document_root=settings.MEDIA_ROOT)

urlpatterns += staticfiles_urlpatterns()

# API URLS

urlpatterns += [
    # API base url
    path("api/", include("tests.api_router")),
    # DRF auth token
    path("auth-token/", obtain_auth_token),
    path("api/schema/", SpectacularAPIView.as_view(), name="api-schema"),
    path(
        "api/docs/",
        SpectacularSwaggerView.as_view(url_name="api-schema"),
        name="api-docs",
    ),
]



"""Defines signals

quote_random_retrieved is emitted when a random quote is supplied.

The `sender` should in most contexts be either `Source` or `SourceGroup` class definitions (**not instances**).
The `instance` should be the actual instance of the `Source` that is being used.

This signal will update the ``quotes_retrieved`` stats in the related ``GroupStats``, ``SourceStats``, and ``QuoteStats`` objects.
"""

import django.dispatch

quote_random_retrieved = django.dispatch.Signal()



import pytest
from django_quotes import views


# Testing for documented views per: https://simonwillison.net/2018/Jul/28/documentation-unit-tests/
@pytest.mark.xfail
@pytest.mark.parametrize("view_class", [v for v in dir(views) if v.endswith("View")])
def test_view_classes_are_documented(documented_views, view_class):
    assert view_class in documented_views



# Generated by Django 5.0.4 on 2024-04-03 20:41

import django.db.models.deletion
import rules.contrib.models
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    replaces = [
        ("django_quotes", "0001_initial"),
        ("django_quotes", "0002_alter_sourcegroup_name"),
        ("django_quotes", "0003_alter_source_slug_alter_sourcegroup_slug"),
        ("django_quotes", "0004_quote_pub_date"),
        ("django_quotes", "0005_alter_groupmarkovmodel_created_and_more"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="SourceGroup",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                (
                    "public",
                    models.BooleanField(
                        default=False, help_text="Is this a public source available for any user to view?"
                    ),
                ),
                (
                    "allow_submissions",
                    models.BooleanField(default=False, help_text="Allow submissions from other users?"),
                ),
                (
                    "name",
                    models.CharField(
                        db_index=True,
                        help_text="A source for individuals making the quotes. Use as an abstract grouping.",
                        max_length=50,
                        verbose_name="Source Name",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True, help_text="Description for the source. You can style using Markdown.", null=True
                    ),
                ),
                (
                    "description_rendered",
                    models.TextField(blank=True, help_text="Automatically generated from description", null=True),
                ),
                (
                    "slug",
                    models.SlugField(
                        blank=True, editable=False, help_text="Unique slug for this group.", max_length=70, unique=True
                    ),
                ),
                ("owner", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                "ordering": ["name"],
            },
            bases=(rules.contrib.models.RulesModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="Source",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                (
                    "public",
                    models.BooleanField(
                        default=False, help_text="Is this a public source available for any user to view?"
                    ),
                ),
                (
                    "allow_submissions",
                    models.BooleanField(default=False, help_text="Allow submissions from other users?"),
                ),
                ("name", models.CharField(help_text="Name of the character", max_length=100)),
                (
                    "slug",
                    models.SlugField(
                        blank=True,
                        editable=False,
                        help_text="Global slug of the character, will be auto generated from name and group if not overridden.",
                        max_length=250,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Description of this character. You can style this with Markdown.",
                        null=True,
                    ),
                ),
                (
                    "description_rendered",
                    models.TextField(blank=True, help_text="Automatically generated from description.", null=True),
                ),
                ("allow_markov", models.BooleanField(default=False, help_text="Allow to be used in markov chains?")),
                (
                    "group",
                    models.ForeignKey(
                        help_text="The group this character belongs to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_quotes.sourcegroup",
                    ),
                ),
                ("owner", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            bases=(rules.contrib.models.RulesModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="GroupMarkovModel",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                ("data", models.JSONField(blank=True, help_text="The cached markov model as JSON.", null=True)),
                (
                    "group",
                    models.OneToOneField(
                        help_text="The character group this model belongs to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_quotes.sourcegroup",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="GroupStats",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                (
                    "quotes_requested",
                    models.PositiveIntegerField(
                        default=0, help_text="Number of time child quotes have been requested."
                    ),
                ),
                (
                    "quotes_generated",
                    models.PositiveIntegerField(
                        default=0, help_text="Number of times markov generated quotes have been requested."
                    ),
                ),
                (
                    "group",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stats",
                        to="django_quotes.sourcegroup",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Quote",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                (
                    "public",
                    models.BooleanField(
                        default=False, help_text="Is this a public source available for any user to view?"
                    ),
                ),
                (
                    "allow_submissions",
                    models.BooleanField(default=False, help_text="Allow submissions from other users?"),
                ),
                (
                    "quote",
                    models.CharField(
                        help_text="Plain text representation of quote. You can use Markdown here.", max_length=280
                    ),
                ),
                (
                    "quote_rendered",
                    models.TextField(
                        blank=True,
                        help_text="HTML rendered version of quote generated from quote plain text.",
                        null=True,
                    ),
                ),
                (
                    "citation",
                    models.CharField(
                        blank=True, help_text="Where is this quote from? Episode #, book?", max_length=250, null=True
                    ),
                ),
                ("citation_url", models.URLField(blank=True, help_text="URL for citation, if applicable.", null=True)),
                ("owner", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
                (
                    "source",
                    models.ForeignKey(
                        help_text="The source of this quote, i.e. name.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="django_quotes.source",
                    ),
                ),
                (
                    "pub_date",
                    models.DateTimeField(
                        blank=True,
                        help_text="When is the earliest time this should appear in random results?",
                        null=True,
                    ),
                ),
            ],
            bases=(rules.contrib.models.RulesModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="QuoteStats",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                ("times_used", models.PositiveIntegerField(default=0, help_text="Times used for random quotes, etc.")),
                (
                    "quote",
                    models.OneToOneField(
                        help_text="The Quote the stats related to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="stats",
                        to="django_quotes.quote",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="SourceMarkovModel",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                ("data", models.JSONField(blank=True, null=True)),
                (
                    "source",
                    models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to="django_quotes.source"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="SourceStats",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created", models.DateTimeField(auto_now_add=True, help_text="Created timestamp")),
                ("modified", models.DateTimeField(auto_now=True, help_text="Last modified time")),
                (
                    "quotes_requested",
                    models.PositiveIntegerField(
                        default=0, help_text="Number of time child quotes have been requested."
                    ),
                ),
                (
                    "quotes_generated",
                    models.PositiveIntegerField(
                        default=0, help_text="Number of times markov generated quotes have been requested."
                    ),
                ),
                (
                    "source",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE, related_name="stats", to="django_quotes.source"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models import F
from django.db.models.signals import post_save, pre_save
from django.dispatch import receiver
from markdown import markdown

from django_markov.models import MarkovTextModel, sentence_generated
from django_quotes.models import (
    GroupStats,
    Quote,
    QuoteStats,
    Source,
    SourceGroup,
    SourceStats,
)
from django_quotes.signals import quote_random_retrieved


@receiver(pre_save, sender=SourceGroup)
@receiver(pre_save, sender=Source)
def render_description(sender, instance, *args, **kwargs):
    """
    Automatically renders the description from markdown.
    """
    if instance.description:
        instance.description_rendered = markdown(instance.description)
    else:
        instance.description_rendered = None


@receiver(pre_save, sender=Quote)
def render_quote(sender, instance, *args, **kwargs):
    """
    Render the quote via markdown and save the results.
    """
    instance.quote_rendered = markdown(instance.quote)


@receiver(pre_save, sender=SourceGroup)
def initialize_group_markov_object(sender, instance, *args, **kwargs):
    """
    Creates the one-to-one object for the group markov model.
    """
    if not instance.text_model:
        instance.text_model = MarkovTextModel.objects.create()


@receiver(pre_save, sender=Source)
def initialize_markov_object(sender, instance, *args, **kwargs):
    """
    Creates the one-to-one object to accompany the source object.
    """
    if not instance.text_model:
        instance.text_model = MarkovTextModel.objects.create()


@receiver(post_save, sender=SourceGroup)
@receiver(post_save, sender=Source)
@receiver(post_save, sender=Quote)
def initialize_grouping_stat_object(sender, instance, created, *args, **kwargs):
    """
    Creates the initial stat objects in the database.
    """
    if created:
        if sender == SourceGroup:
            GroupStats.objects.create(group=instance)
        elif sender == Source:
            SourceStats.objects.create(source=instance)
        elif sender == Quote:
            QuoteStats.objects.create(quote=instance)


@receiver(quote_random_retrieved, sender=Source)
def update_stats_for_quote_character(sender, instance, quote_retrieved, *args, **kwargs):
    """
    Update the stats for the source, source group, and quote for a random retrieval.
    :param sender: Usually a source or sourcegroup class.
    :param instance: The source this was generated for.
    :param quote_retrieved: The quote that was returned.
    :return: None
    """
    group_stats = instance.group.stats
    character_stats = instance.stats
    quote_stats = quote_retrieved.stats
    with transaction.atomic():
        group_stats.quotes_requested = F("quotes_requested") + 1
        group_stats.save()
        character_stats.quotes_requested = F("quotes_requested") + 1
        character_stats.save()
        quote_stats.times_used = F("times_used") + 1
        quote_stats.save()


@receiver(sentence_generated, sender=MarkovTextModel)
def update_stats_for_markov(sender, instance, char_limit, sentence, *args, **kwargs):
    """
    For a given source, update the stats on the Source and SourceGroup for markov requests.
    :param sender: The requesting class, usually Source.
    :param instance: The specific source requested.
    :param char_limit: The character limit used when generating the sentence.
    :param sentence: The sentence that was generated.
    :return: None
    """
    try:
        source = Source.objects.get(text_model__pk=instance.pk)
        group = source.group
    except ObjectDoesNotExist:
        # this is a group model
        source = None
        group = SourceGroup.objects.get(text_model__pk=instance.pk)
    group_stats = group.stats
    source_stats = source.stats if source else None
    with transaction.atomic():
        group_stats.quotes_generated = F("quotes_generated") + 1
        group_stats.save()
        if source_stats is not None:
            source_stats.quotes_generated = F("quotes_generated") + 1
            source_stats.save()


# @receiver(post_save, sender=Quote)
# def update_markov_model_on_quote_change(sender, instance, created, *args, **kwargs):
#     if instance.character.allow_markov:
#         cmm = CharacterMarkovModel.objects.get(character=instance.character)
#         cmm.generate_model_from_corpus()


@receiver(pre_save, sender=Source)
def update_markov_model_for_character_enabling_markov(sender, instance, *args, **kwargs):
    if instance.id and instance.allow_markov:
        old_version = Source.objects.get(id=instance.id)
        if not old_version.allow_markov:
            instance.update_markov_model()
            instance.group.update_markov_model()



# Generated by Django 5.0.4 on 2024-04-05 20:30

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("django_quotes", "0011_alter_source_text_model_alter_sourcegroup_text_model"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="sourcemarkovmodel",
            name="source",
        ),
        migrations.DeleteModel(
            name="GroupMarkovModel",
        ),
        migrations.DeleteModel(
            name="SourceMarkovModel",
        ),
    ]



from django.core.exceptions import ObjectDoesNotExist
from drf_spectacular.utils import extend_schema, inline_serializer
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.fields import CharField
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet
from rules.contrib.rest_framework import AutoPermissionViewSetMixin

from django_quotes.api.serializers import (
    QuoteSerializer,
    SourceGroupSerializer,
    SourceSerializer,
)
from django_quotes.models import Source, SourceGroup


class SourceGroupViewSet(AutoPermissionViewSetMixin, RetrieveModelMixin, ListModelMixin, GenericViewSet):
    """
    A generic viewset for listing and retrieving details on sourceGroup groups.
    """

    serializer_class = SourceGroupSerializer
    lookup_field = "slug"
    lookup_url_kwarg = "group"
    permission_type_map = {
        "create": "add",
        "destroy": "delete",
        "list": None,
        "partial_update": "change",
        "retrieve": "read",
        "update": "change",
        "get_random_quote": "read",
        "generate_sentence": "read",
    }

    def get_queryset(self, *args, **kwargs):
        return SourceGroup.objects.filter(owner=self.request.user) | SourceGroup.objects.filter(  # type: ignore
            public=True
        )

    @extend_schema(responses={200: QuoteSerializer})
    @action(detail=True, methods=["get"])
    def get_random_quote(self, request, group=None):
        g = self.get_object()
        quote = g.get_random_quote()
        if quote is not None:
            qs = QuoteSerializer(quote)
            return Response(status=status.HTTP_200_OK, data=qs.data)
        return Response(status=status.HTTP_404_NOT_FOUND, data={"error": "No quotes found."})

    @extend_schema(responses={200: inline_serializer(name="generated_sentence", fields={"sentence": CharField()})})
    @action(detail=True, methods=["get"])
    def generate_sentence(self, request, group=None):
        g = self.get_object()
        if g.markov_sources == 0:
            return Response(
                status=status.HTTP_403_FORBIDDEN,
                data={"error": "This group does not currently allow sentence generation."},
            )
        sentence = g.generate_markov_sentence()
        if sentence is not None:
            return Response(status=status.HTTP_200_OK, data={"sentence": sentence})
        return Response(
            status=status.HTTP_204_NO_CONTENT,
            data={"error": "Insufficent data to generate sentence."},
        )


class SourceViewSet(AutoPermissionViewSetMixin, RetrieveModelMixin, ListModelMixin, GenericViewSet):
    """
    Retrieve and list views for sources.
    """

    serializer_class = SourceSerializer
    lookup_field = "slug"
    lookup_url_kwarg = "source"
    permission_type_map = {
        "create": "add",
        "destroy": "delete",
        "list": None,
        "partial_update": "change",
        "retrieve": "read",
        "update": "change",
        "get_random_quote": "read",
        "generate_sentence": "read",
    }

    def get_queryset(self, *args, **kwargs):
        group_slug = self.request.query_params.get("group")  # type: ignore
        queryset = Source.objects.filter(owner=self.request.user) | Source.objects.filter(public=True)  # type: ignore
        if group_slug:
            try:
                group = SourceGroup.objects.get(slug=group_slug)
            except ObjectDoesNotExist:
                return SourceGroup.objects.none()
            queryset = queryset.filter(group=group)
        return queryset

    @extend_schema(responses={200: QuoteSerializer})
    @action(detail=True, methods=["get"])
    def get_random_quote(self, request, source=None):
        source = self.get_object()
        quote = source.get_random_quote()
        if quote is not None:
            qs = QuoteSerializer(quote)
            return Response(status=status.HTTP_200_OK, data=qs.data)
        return Response(status=status.HTTP_404_NOT_FOUND, data={"error": "No quotes found."})

    @extend_schema(responses={200: inline_serializer(name="generated_sentence", fields={"sentence": CharField()})})
    @action(detail=True, methods=["get"])
    def generate_sentence(self, request, source=None):
        source = self.get_object()
        if not source.allow_markov:
            return Response(
                status=status.HTTP_403_FORBIDDEN,
                data={"error": "This source does not permit sentence generation."},
            )
        sentence = source.get_markov_sentence()
        if sentence is not None:
            return Response(status=status.HTTP_200_OK, data={"sentence": sentence})
        return Response(
            status=status.HTTP_204_NO_CONTENT,
            data={"error": "Unable to generate markov sentence. This source may not have enough quotes yet."},
        )



from datetime import timedelta

import pytest
from django.contrib.auth import get_user_model
from django.db import IntegrityError
from django.utils import timezone
from django_quotes.models import Quote, Source, SourceGroup

User = get_user_model()

pytestmark = pytest.mark.django_db(transaction=True)


def test_generate_group_slug(user: User) -> None:
    """
    Test the slug generation for source groups.
    :param user: The logged in user.
    :return:
    """
    group = SourceGroup.objects.create(name="Curious George", owner=user)
    assert group.slug == "curious-george"


def test_generate_source_slug(user: User) -> None:
    """
    Tests the slug generation for sources to ensure it's being set up correctly.

    :param user: The user who is currently logged in.
    :return:
    """
    group = SourceGroup.objects.create(name="Monkey", owner=user)
    source = Source.objects.create(name="John Smith", group=group, owner=user)
    assert source.slug == "monkey-john-smith"


@pytest.mark.parametrize(
    "group1,group2,source_name,slug",
    [
        ("EW", "Explorers Wanted", "John Smith", "ew-john-smith"),
    ],
)
def test_reject_source_duplicate_slug(user: User, group1: str, group2: str, source_name: str, slug: str) -> None:
    """
    :param group1: Name of the first group
    :param group2: Name of the second group
    :param source_name: Name of the source.
    :param slug: Slug to override.
    :param user: The user who is currently logged in.
    :return:
    """
    group = SourceGroup.objects.create(name=group1, owner=user)
    Source.objects.create(name=source_name, group=group, owner=user)
    group2 = SourceGroup.objects.create(name=group2, owner=user)
    with pytest.raises(IntegrityError):
        Source.objects.create(name=source_name, group=group2, slug=slug, owner=user)


def test_group_properties_calculation(property_group: SourceGroup) -> None:
    assert property_group.total_sources == 10
    assert property_group.markov_sources == 5
    assert property_group.total_quotes == 200


def test_refresh_from_db_also_updates_cached_properties(property_group: SourceGroup, user: User) -> None:
    assert property_group.total_sources == 10
    assert property_group.markov_sources == 5
    assert property_group.total_quotes == 200
    c = Source.objects.create(name="IamNew", group=property_group, allow_markov=True, owner=user)
    Quote.objects.create(source=c, quote="I'm a new quote", owner=user)
    assert property_group.total_sources == 10
    assert property_group.markov_sources == 5
    assert property_group.total_quotes == 200
    property_group.refresh_from_db()
    assert property_group.total_sources == 11
    assert property_group.markov_sources == 6
    assert property_group.total_quotes == 201


def test_retrieve_random_quote(property_group):
    noquote_source = Source.objects.create(group=property_group, name="No One", owner=property_group.owner)
    assert noquote_source.get_random_quote() is None
    noquote_source.delete()
    quoteable_source = Source.objects.filter(group=property_group)[0]
    assert type(quoteable_source.get_random_quote()) == Quote


def test_generate_markov_sentence(property_group):
    noquote_source = Source.objects.create(group=property_group, name="No One", owner=property_group.owner)
    assert noquote_source.get_markov_sentence() is None
    noquote_source.delete()
    quotable_source = Source.objects.filter(group=property_group)[0]
    sentence = quotable_source.get_markov_sentence(tries=50)
    print(sentence)
    assert isinstance(sentence, str)


def test_get_random_group_quote(property_group):
    noquote_group = SourceGroup.objects.create(name="I am no one.", owner=property_group.owner)
    assert noquote_group.get_random_quote() is None
    assert isinstance(property_group.get_random_quote(), Quote)


def test_group_generate_markov_sentence(property_group, corpus_sentences):
    no_quote_group = SourceGroup.objects.create(name="We are no one.", owner=property_group.owner)
    Source.objects.create(
        name="John Doe",
        group=no_quote_group,
        allow_markov=True,
        owner=property_group.owner,
    )
    quote_source = Source.objects.create(
        name="Jane Doe",
        group=no_quote_group,
        allow_markov=False,
        owner=property_group.owner,
    )
    for sentence in corpus_sentences:
        Quote.objects.create(source=quote_source, quote=sentence, owner=property_group.owner)
    assert no_quote_group.generate_markov_sentence() is None
    assert property_group.generate_markov_sentence(tries=50) is not None


def test_pub_date_prevents_inclusion_in_source_random_quote(property_group):
    new_source = Source.objects.create(group=property_group, name="Future Man", owner=property_group.owner)
    q = Quote.objects.create(
        source=new_source,
        quote="This message is from the **future**!",
        pub_date=timezone.now() + timedelta(days=4),
        owner=property_group.owner,
    )
    assert Quote.objects.filter(source=new_source).count() == 1
    assert new_source.get_random_quote() is None
    q.pub_date = timezone.now() - timedelta(days=2)
    q.save()
    assert new_source.get_random_quote() == q


def test_pub_date_prevents_inclusion_in_group_random_quote(property_group):
    new_group = SourceGroup.objects.create(name="Future Group", owner=property_group.owner)
    new_source = Source.objects.create(group=new_group, owner=new_group.owner, name="Future Man")
    q = Quote.objects.create(
        source=new_source,
        quote="This message is from the **future**!",
        pub_date=timezone.now() + timedelta(days=4),
        owner=property_group.owner,
    )
    assert Quote.objects.filter(source__group=new_group).count() == 1
    assert new_group.get_random_quote() is None
    q.pub_date = timezone.now() - timedelta(days=2)
    q.save()
    assert new_group.get_random_quote() == q
